@@ -27,11 +27,12 @@ import (
 	"golang.org/x/tools/gopls/internal/lsprpc"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/protocol/command"
+	"golang.org/x/tools/gopls/internal/protocol/semtok"
 	"golang.org/x/tools/gopls/internal/server"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/browser"
 	bugpkg "golang.org/x/tools/gopls/internal/util/bug"
-	"golang.org/x/tools/gopls/internal/util/constraints"
+	"golang.org/x/tools/gopls/internal/util/moreslices"
 	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/jsonrpc2"
 	"golang.org/x/tools/internal/tool"
@@ -62,9 +63,6 @@ type Application struct {
 	// VeryVerbose enables a higher level of verbosity in logging output.
 	VeryVerbose bool `flag:"vv,veryverbose" help:"very verbose output"`
 
-	// Control ocagent export of telemetry
-	OCAgent string `flag:"ocagent" help:"the address of the ocagent (e.g. http://localhost:55678), or off"`
-
 	// PrepareOptions is called to update the options when a new view is built.
 	// It is primarily to allow the behavior of gopls to be modified by hooks.
 	PrepareOptions func(*settings.Options)
@@ -75,7 +73,7 @@ type Application struct {
 	editFlags *EditFlags
 }
 
-// EditFlags defines flags common to {fix,format,imports,rename}
+// EditFlags defines flags common to {code{action,lens},format,imports,rename}
 // that control how edits are applied to the client's files.
 //
 // The type is exported for flag reflection.
@@ -95,11 +93,8 @@ func (app *Application) verbose() bool {
 }
 
 // New returns a new Application ready to run.
-func New(options func(*settings.Options)) *Application {
+func New() *Application {
 	app := &Application{
-		options: options,
-		OCAgent: "off", //TODO: Remove this line to default the exporter to on
-
 		Serve: Serve{
 			RemoteListenTimeout: 1 * time.Minute,
 		},
@@ -132,6 +127,10 @@ It is typically used with an editor to provide language features. When no
 command is specified, gopls will default to the 'serve' command. The language
 features can also be accessed via the gopls command-line interface.
 
+For documentation of all its features, see:
+
+   https://github.com/golang/tools/blob/master/gopls/doc/features
+
 Usage:
   gopls help [<subject>]
 
@@ -216,7 +215,7 @@ func isZeroValue(f *flag.Flag, value string) bool {
 	// This works unless the Value type is itself an interface type.
 	typ := reflect.TypeOf(f.Value)
 	var z reflect.Value
-	if typ.Kind() == reflect.Ptr {
+	if typ.Kind() == reflect.Pointer {
 		z = reflect.New(typ.Elem())
 	} else {
 		z = reflect.Zero(typ)
@@ -234,7 +233,7 @@ func (app *Application) Run(ctx context.Context, args ...string) error {
 	// executable, and immediately runs a gc.
 	filecache.Start()
 
-	ctx = debug.WithInstance(ctx, app.OCAgent)
+	ctx = debug.WithInstance(ctx)
 	if len(args) == 0 {
 		s := flag.NewFlagSet(app.Name(), flag.ExitOnError)
 		return tool.Run(ctx, s, &app.Serve, args)
@@ -280,12 +279,15 @@ func (app *Application) internalCommands() []tool.Application {
 func (app *Application) featureCommands() []tool.Application {
 	return []tool.Application{
 		&callHierarchy{app: app},
-		&check{app: app},
+		&check{app: app, Severity: "warning"},
+		&codeaction{app: app},
 		&codelens{app: app},
 		&definition{app: app},
 		&execute{app: app},
+		&fix{app: app}, // (non-functional)
 		&foldingRanges{app: app},
 		&format{app: app},
+		&headlessMCP{app: app},
 		&highlight{app: app},
 		&implementation{app: app},
 		&imports{app: app},
@@ -295,70 +297,59 @@ func (app *Application) featureCommands() []tool.Application {
 		&prepareRename{app: app},
 		&references{app: app},
 		&rename{app: app},
-		&semtok{app: app},
+		&semanticToken{app: app},
 		&signature{app: app},
 		&stats{app: app},
-		&suggestedFix{app: app},
 		&symbols{app: app},
 
 		&workspaceSymbol{app: app},
 	}
 }
 
-var (
-	internalMu          sync.Mutex
-	internalConnections = make(map[string]*connection)
-)
-
-// connect creates and initializes a new in-process gopls session.
-//
-// If onProgress is set, it is called for each new progress notification.
-func (app *Application) connect(ctx context.Context, onProgress func(*protocol.ProgressParams)) (*connection, error) {
-	switch {
-	case app.Remote == "":
-		client := newClient(app, onProgress)
-		options := settings.DefaultOptions(app.options)
-		server := server.New(cache.NewSession(ctx, cache.New(nil)), client, options)
-		conn := newConnection(server, client)
-		if err := conn.initialize(protocol.WithClient(ctx, client), app.options); err != nil {
-			return nil, err
+// connect creates and initializes a new in-process gopls LSP session.
+func (app *Application) connect(ctx context.Context) (*client, *cache.Session, error) {
+	root, err := os.Getwd()
+	if err != nil {
+		return nil, nil, fmt.Errorf("finding workdir: %v", err)
+	}
+	options := settings.DefaultOptions(app.options)
+	client := newClient(app)
+	var (
+		svr  protocol.Server
+		sess *cache.Session
+	)
+	if app.Remote == "" {
+		// local
+		sess = cache.NewSession(ctx, cache.New(nil))
+		svr = server.New(sess, client, options)
+		ctx = protocol.WithClient(ctx, client)
+	} else {
+		// remote
+		netConn, err := lsprpc.ConnectToRemote(ctx, app.Remote)
+		if err != nil {
+			return nil, nil, err
 		}
-		return conn, nil
-
-	default:
-		return app.connectRemote(ctx, app.Remote)
+		stream := jsonrpc2.NewHeaderStream(netConn)
+		jsonConn := jsonrpc2.NewConn(stream)
+		svr = protocol.ServerDispatcher(jsonConn)
+		ctx = protocol.WithClient(ctx, client)
+		jsonConn.Go(ctx,
+			protocol.Handlers(
+				protocol.ClientHandler(client, jsonrpc2.MethodNotFound)))
 	}
+	if err := client.initialize(ctx, svr, initParams(root, options)); err != nil {
+		return nil, nil, err
+	}
+	return client, sess, nil
 }
 
-func (app *Application) connectRemote(ctx context.Context, remote string) (*connection, error) {
-	conn, err := lsprpc.ConnectToRemote(ctx, remote)
-	if err != nil {
-		return nil, err
-	}
-	stream := jsonrpc2.NewHeaderStream(conn)
-	cc := jsonrpc2.NewConn(stream)
-	server := protocol.ServerDispatcher(cc)
-	client := newClient(app, nil)
-	connection := newConnection(server, client)
-	ctx = protocol.WithClient(ctx, connection.client)
-	cc.Go(ctx,
-		protocol.Handlers(
-			protocol.ClientHandler(client, jsonrpc2.MethodNotFound)))
-	return connection, connection.initialize(ctx, app.options)
-}
-
-func (c *connection) initialize(ctx context.Context, options func(*settings.Options)) error {
-	wd, err := os.Getwd()
-	if err != nil {
-		return fmt.Errorf("finding workdir: %v", err)
-	}
+func initParams(rootDir string, opts *settings.Options) *protocol.ParamInitialize {
 	params := &protocol.ParamInitialize{}
-	params.RootURI = protocol.URIFromPath(wd)
+	params.RootURI = protocol.URIFromPath(rootDir)
 	params.Capabilities.Workspace.Configuration = true
 
-	// Make sure to respect configured options when sending initialize request.
-	opts := settings.DefaultOptions(options)
-	// If you add an additional option here, you must update the map key in connect.
+	// If you add an additional option here,
+	// you must update the map key of settings.DefaultOptions called in (*Application).connect.
 	params.Capabilities.TextDocument.Hover = &protocol.HoverClientCapabilities{
 		ContentFormat: []protocol.MarkupKind{opts.PreferredContentFormat},
 	}
@@ -366,81 +357,98 @@ func (c *connection) initialize(ctx context.Context, options func(*settings.Opti
 	params.Capabilities.TextDocument.SemanticTokens = protocol.SemanticTokensClientCapabilities{}
 	params.Capabilities.TextDocument.SemanticTokens.Formats = []protocol.TokenFormat{"relative"}
 	params.Capabilities.TextDocument.SemanticTokens.Requests.Range = &protocol.Or_ClientSemanticTokensRequestOptions_range{Value: true}
-	//params.Capabilities.TextDocument.SemanticTokens.Requests.Range.Value = true
+	// params.Capabilities.TextDocument.SemanticTokens.Requests.Range.Value = true
 	params.Capabilities.TextDocument.SemanticTokens.Requests.Full = &protocol.Or_ClientSemanticTokensRequestOptions_full{Value: true}
-	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = protocol.SemanticTypes()
-	params.Capabilities.TextDocument.SemanticTokens.TokenModifiers = protocol.SemanticModifiers()
-
-	// If the subcommand has registered a progress handler, report the progress
-	// capability.
-	if c.client.onProgress != nil {
-		params.Capabilities.Window.WorkDoneProgress = true
+	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = moreslices.ConvertStrings[string](semtok.TokenTypes)
+	params.Capabilities.TextDocument.SemanticTokens.TokenModifiers = moreslices.ConvertStrings[string](semtok.TokenModifiers)
+	params.Capabilities.TextDocument.CodeAction = protocol.CodeActionClientCapabilities{
+		CodeActionLiteralSupport: protocol.ClientCodeActionLiteralOptions{
+			CodeActionKind: protocol.ClientCodeActionKindOptions{
+				ValueSet: []protocol.CodeActionKind{protocol.Empty}, // => all
+			},
+		},
 	}
-
-	params.InitializationOptions = map[string]interface{}{
+	params.Capabilities.Window.WorkDoneProgress = true
+	params.Capabilities.Workspace.FileOperations = &protocol.FileOperationClientCapabilities{
+		DidCreate: true,
+	}
+	params.InitializationOptions = map[string]any{
 		"symbolMatcher": string(opts.SymbolMatcher),
 	}
-	if _, err := c.Server.Initialize(ctx, params); err != nil {
+	return params
+}
+
+// initialize performs LSP's two-call client/server handshake.
+func (cli *client) initialize(ctx context.Context, server protocol.Server, params *protocol.ParamInitialize) error {
+	result, err := server.Initialize(ctx, params)
+	if err != nil {
 		return err
 	}
-	if err := c.Server.Initialized(ctx, &protocol.InitializedParams{}); err != nil {
+	if err := server.Initialized(ctx, &protocol.InitializedParams{}); err != nil {
 		return err
 	}
+	cli.server = server
+	cli.initializeResult = result
 	return nil
 }
 
-type connection struct {
-	protocol.Server
-	client *cmdClient
-}
+// client implements [protocol.Client] and defines the LSP client
+// operations of the gopls command.
+//
+// It holds the client-side state of a single client/server
+// connection; it conceptually corresponds to a single call to
+// connect(2).
+type client struct {
+	app *Application
 
-// cmdClient defines the protocol.Client interface behavior of the gopls CLI tool.
-type cmdClient struct {
-	app        *Application
-	onProgress func(*protocol.ProgressParams)
+	server           protocol.Server
+	initializeResult *protocol.InitializeResult // includes server capabilities
 
-	filesMu sync.Mutex // guards files map and each cmdFile.diagnostics
+	progressMu sync.Mutex
+	iwlToken   protocol.ProgressToken
+	iwlDone    chan struct{}
+
+	filesMu sync.Mutex // guards files map
 	files   map[protocol.DocumentURI]*cmdFile
 }
 
+// cmdFile represents an open file in the gopls command LSP client.
 type cmdFile struct {
-	uri         protocol.DocumentURI
-	mapper      *protocol.Mapper
-	err         error
-	diagnostics []protocol.Diagnostic
+	uri           protocol.DocumentURI
+	mapper        *protocol.Mapper
+	err           error
+	diagnosticsMu sync.Mutex
+	diagnostics   []protocol.Diagnostic
 }
 
-func newClient(app *Application, onProgress func(*protocol.ProgressParams)) *cmdClient {
-	return &cmdClient{
-		app:        app,
-		onProgress: onProgress,
-		files:      make(map[protocol.DocumentURI]*cmdFile),
+func newClient(app *Application) *client {
+	return &client{
+		app:     app,
+		files:   make(map[protocol.DocumentURI]*cmdFile),
+		iwlDone: make(chan struct{}),
 	}
 }
 
-func newConnection(server protocol.Server, client *cmdClient) *connection {
-	return &connection{
-		Server: server,
-		client: client,
-	}
+func (cli *client) TextDocumentContentRefresh(context.Context, *protocol.TextDocumentContentRefreshParams) error {
+	return nil
 }
 
-func (c *cmdClient) CodeLensRefresh(context.Context) error { return nil }
+func (cli *client) CodeLensRefresh(context.Context) error { return nil }
 
-func (c *cmdClient) FoldingRangeRefresh(context.Context) error { return nil }
+func (cli *client) FoldingRangeRefresh(context.Context) error { return nil }
 
-func (c *cmdClient) LogTrace(context.Context, *protocol.LogTraceParams) error { return nil }
+func (cli *client) LogTrace(context.Context, *protocol.LogTraceParams) error { return nil }
 
-func (c *cmdClient) ShowMessage(ctx context.Context, p *protocol.ShowMessageParams) error {
+func (cli *client) ShowMessage(ctx context.Context, p *protocol.ShowMessageParams) error {
 	fmt.Fprintf(os.Stderr, "%s: %s\n", p.Type, p.Message)
 	return nil
 }
 
-func (c *cmdClient) ShowMessageRequest(ctx context.Context, p *protocol.ShowMessageRequestParams) (*protocol.MessageActionItem, error) {
+func (cli *client) ShowMessageRequest(ctx context.Context, p *protocol.ShowMessageRequestParams) (*protocol.MessageActionItem, error) {
 	return nil, nil
 }
 
-func (c *cmdClient) LogMessage(ctx context.Context, p *protocol.LogMessageParams) error {
+func (cli *client) LogMessage(ctx context.Context, p *protocol.LogMessageParams) error {
 	// This logic causes server logging to be double-prefixed with a timestamp.
 	//     2023/11/08 10:50:21 Error:2023/11/08 10:50:21 <actual message>
 	// TODO(adonovan): print just p.Message, plus a newline if needed?
@@ -450,42 +458,42 @@ func (c *cmdClient) LogMessage(ctx context.Context, p *protocol.LogMessageParams
 	case protocol.Warning:
 		log.Print("Warning:", p.Message)
 	case protocol.Info:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print("Info:", p.Message)
 		}
 	case protocol.Log:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print("Log:", p.Message)
 		}
 	default:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print(p.Message)
 		}
 	}
 	return nil
 }
 
-func (c *cmdClient) Event(ctx context.Context, t *interface{}) error { return nil }
+func (cli *client) Event(ctx context.Context, t *any) error { return nil }
 
-func (c *cmdClient) RegisterCapability(ctx context.Context, p *protocol.RegistrationParams) error {
+func (cli *client) RegisterCapability(ctx context.Context, p *protocol.RegistrationParams) error {
 	return nil
 }
 
-func (c *cmdClient) UnregisterCapability(ctx context.Context, p *protocol.UnregistrationParams) error {
+func (cli *client) UnregisterCapability(ctx context.Context, p *protocol.UnregistrationParams) error {
 	return nil
 }
 
-func (c *cmdClient) WorkspaceFolders(ctx context.Context) ([]protocol.WorkspaceFolder, error) {
+func (cli *client) WorkspaceFolders(ctx context.Context) ([]protocol.WorkspaceFolder, error) {
 	return nil, nil
 }
 
-func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfiguration) ([]interface{}, error) {
-	results := make([]interface{}, len(p.Items))
+func (cli *client) Configuration(ctx context.Context, p *protocol.ParamConfiguration) ([]any, error) {
+	results := make([]any, len(p.Items))
 	for i, item := range p.Items {
 		if item.Section != "gopls" {
 			continue
 		}
-		m := map[string]interface{}{
+		m := map[string]any{
 			"analyses": map[string]any{
 				"fillreturns":    true,
 				"nonewvars":      true,
@@ -493,7 +501,7 @@ func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfigur
 				"undeclaredname": true,
 			},
 		}
-		if c.app.VeryVerbose {
+		if cli.app.VeryVerbose {
 			m["verboseOutput"] = true
 		}
 		results[i] = m
@@ -501,8 +509,8 @@ func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfigur
 	return results, nil
 }
 
-func (c *cmdClient) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEditParams) (*protocol.ApplyWorkspaceEditResult, error) {
-	if err := c.applyWorkspaceEdit(&p.Edit); err != nil {
+func (cli *client) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEditParams) (*protocol.ApplyWorkspaceEditResult, error) {
+	if err := cli.applyWorkspaceEdit(&p.Edit); err != nil {
 		return &protocol.ApplyWorkspaceEditResult{FailureReason: err.Error()}, nil
 	}
 	return &protocol.ApplyWorkspaceEditResult{Applied: true}, nil
@@ -511,68 +519,115 @@ func (c *cmdClient) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEdi
 // applyWorkspaceEdit applies a complete WorkspaceEdit to the client's
 // files, honoring the preferred edit mode specified by cli.app.editMode.
 // (Used by rename and by ApplyEdit downcalls.)
-func (cli *cmdClient) applyWorkspaceEdit(edit *protocol.WorkspaceEdit) error {
-	var orderedURIs []protocol.DocumentURI
-	edits := map[protocol.DocumentURI][]protocol.TextEdit{}
-	for _, c := range edit.DocumentChanges {
-		if c.TextDocumentEdit != nil {
-			uri := c.TextDocumentEdit.TextDocument.URI
-			edits[uri] = append(edits[uri], protocol.AsTextEdits(c.TextDocumentEdit.Edits)...)
-			orderedURIs = append(orderedURIs, uri)
-		}
-		if c.RenameFile != nil {
-			return fmt.Errorf("client does not support file renaming (%s -> %s)",
-				c.RenameFile.OldURI,
-				c.RenameFile.NewURI)
-		}
+//
+// See also:
+//   - changedFiles in ../test/marker/marker_test.go for the golden-file capturing variant
+//   - applyWorkspaceEdit in ../test/integration/fake/editor.go for the Editor variant
+func (cli *client) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
+
+	create := func(uri protocol.DocumentURI, content []byte) error {
+		edits := []diff.Edit{{Start: 0, End: 0, New: string(content)}}
+		return updateFile(uri.Path(), nil, content, edits, cli.app.editFlags)
 	}
-	sortSlice(orderedURIs)
-	for _, uri := range orderedURIs {
-		f := cli.openFile(uri)
-		if f.err != nil {
-			return f.err
-		}
-		if err := applyTextEdits(f.mapper, edits[uri], cli.app.editFlags); err != nil {
-			return err
+
+	delete := func(uri protocol.DocumentURI, content []byte) error {
+		edits := []diff.Edit{{Start: 0, End: len(content), New: ""}}
+		return updateFile(uri.Path(), content, nil, edits, cli.app.editFlags)
+	}
+
+	for _, c := range wsedit.DocumentChanges {
+		switch {
+		case c.TextDocumentEdit != nil:
+			f := cli.getFile(c.TextDocumentEdit.TextDocument.URI)
+			if f.err != nil {
+				return f.err
+			}
+			// TODO(adonovan): sanity-check c.TextDocumentEdit.TextDocument.Version
+			edits := protocol.AsTextEdits(c.TextDocumentEdit.Edits)
+			if err := applyTextEdits(f.mapper, edits, cli.app.editFlags); err != nil {
+				return err
+			}
+
+		case c.CreateFile != nil:
+			if err := create(c.CreateFile.URI, []byte{}); err != nil {
+				return err
+			}
+
+		case c.RenameFile != nil:
+			// Analyze as creation + deletion. (NB: loses file mode.)
+			f := cli.getFile(c.RenameFile.OldURI)
+			if f.err != nil {
+				return f.err
+			}
+			if err := create(c.RenameFile.NewURI, f.mapper.Content); err != nil {
+				return err
+			}
+			if err := delete(f.mapper.URI, f.mapper.Content); err != nil {
+				return err
+			}
+
+		case c.DeleteFile != nil:
+			f := cli.getFile(c.DeleteFile.URI)
+			if f.err != nil {
+				return f.err
+			}
+			if err := delete(f.mapper.URI, f.mapper.Content); err != nil {
+				return err
+			}
+
+		default:
+			return fmt.Errorf("unknown DocumentChange: %#v", c)
 		}
 	}
 	return nil
 }
 
-func sortSlice[T constraints.Ordered](slice []T) {
-	sort.Slice(slice, func(i, j int) bool { return slice[i] < slice[j] })
-}
-
 // applyTextEdits applies a list of edits to the mapper file content,
 // using the preferred edit mode. It is a no-op if there are no edits.
 func applyTextEdits(mapper *protocol.Mapper, edits []protocol.TextEdit, flags *EditFlags) error {
 	if len(edits) == 0 {
 		return nil
 	}
-	newContent, renameEdits, err := protocol.ApplyEdits(mapper, edits)
+	newContent, diffEdits, err := protocol.ApplyEdits(mapper, edits)
 	if err != nil {
 		return err
 	}
+	return updateFile(mapper.URI.Path(), mapper.Content, newContent, diffEdits, flags)
+}
 
-	filename := mapper.URI.Path()
-
+// updateFile performs a content update operation on the specified file.
+// If the old content is nil, the operation creates the file.
+// If the new content is nil, the operation deletes the file.
+// The flags control whether the operation is written, or merely listed, diffed, or printed.
+func updateFile(filename string, old, new []byte, edits []diff.Edit, flags *EditFlags) error {
 	if flags.List {
 		fmt.Println(filename)
 	}
 
 	if flags.Write {
-		if flags.Preserve {
-			if err := os.Rename(filename, filename+".orig"); err != nil {
+		if flags.Preserve && old != nil { // edit or delete
+			if err := os.WriteFile(filename+".orig", old, 0666); err != nil {
 				return err
 			}
 		}
-		if err := os.WriteFile(filename, newContent, 0644); err != nil {
-			return err
+
+		if new != nil {
+			// create or edit
+			if err := os.WriteFile(filename, new, 0666); err != nil {
+				return err
+			}
+		} else {
+			// delete
+			if err := os.Remove(filename); err != nil {
+				return err
+			}
 		}
 	}
 
 	if flags.Diff {
-		unified, err := diff.ToUnified(filename+".orig", filename, string(mapper.Content), renameEdits, diff.DefaultContextLines)
+		// For diffing, creations and deletions are equivalent
+		// updating an empty file and making an existing file empty.
+		unified, err := diff.ToUnified(filename+".orig", filename, string(old), edits, diff.DefaultContextLines)
 		if err != nil {
 			return err
 		}
@@ -580,31 +635,38 @@ func applyTextEdits(mapper *protocol.Mapper, edits []protocol.TextEdit, flags *E
 	}
 
 	// No flags: just print edited file content.
-	// TODO(adonovan): how is this ever useful with multiple files?
+	//
+	// This makes no sense for multiple files.
+	// (We should probably change the default to -diff.)
 	if !(flags.List || flags.Write || flags.Diff) {
-		os.Stdout.Write(newContent)
+		os.Stdout.Write(new)
 	}
 
 	return nil
 }
 
-func (c *cmdClient) PublishDiagnostics(ctx context.Context, p *protocol.PublishDiagnosticsParams) error {
+func (cli *client) PublishDiagnostics(ctx context.Context, p *protocol.PublishDiagnosticsParams) error {
 	// Don't worry about diagnostics without versions.
+	//
+	// (Note: the representation of PublishDiagnosticsParams
+	// cannot distinguish a missing Version from v0, but the
+	// server never sends back an explicit zero.)
 	if p.Version == 0 {
 		return nil
 	}
 
-	c.filesMu.Lock()
-	defer c.filesMu.Unlock()
+	file := cli.getFile(p.URI)
+
+	file.diagnosticsMu.Lock()
+	defer file.diagnosticsMu.Unlock()
 
-	file := c.getFile(p.URI)
 	file.diagnostics = append(file.diagnostics, p.Diagnostics...)
 
 	// Perform a crude in-place deduplication.
 	// TODO(golang/go#60122): replace the gopls.diagnose_files
 	// command with support for textDocument/diagnostic,
 	// so that we don't need to do this de-duplication.
-	type key [6]interface{}
+	type key [6]any
 	seen := make(map[key]bool)
 	out := file.diagnostics[:0]
 	for _, d := range file.diagnostics {
@@ -623,21 +685,44 @@ func (c *cmdClient) PublishDiagnostics(ctx context.Context, p *protocol.PublishD
 	return nil
 }
 
-func (c *cmdClient) Progress(_ context.Context, params *protocol.ProgressParams) error {
-	if c.onProgress != nil {
-		c.onProgress(params)
+func (cli *client) Progress(_ context.Context, params *protocol.ProgressParams) error {
+	if _, ok := params.Token.(string); !ok {
+		return fmt.Errorf("unexpected progress token: %[1]T %[1]v", params.Token)
+	}
+
+	switch v := params.Value.(type) {
+	case *protocol.WorkDoneProgressBegin:
+		if v.Title == server.DiagnosticWorkTitle(server.FromInitialWorkspaceLoad) {
+			cli.progressMu.Lock()
+			cli.iwlToken = params.Token
+			cli.progressMu.Unlock()
+		}
+
+	case *protocol.WorkDoneProgressReport:
+		if cli.app.Verbose {
+			fmt.Fprintln(os.Stderr, v.Message)
+		}
+
+	case *protocol.WorkDoneProgressEnd:
+		cli.progressMu.Lock()
+		iwlToken := cli.iwlToken
+		cli.progressMu.Unlock()
+
+		if params.Token == iwlToken {
+			close(cli.iwlDone)
+		}
 	}
 	return nil
 }
 
-func (c *cmdClient) ShowDocument(ctx context.Context, params *protocol.ShowDocumentParams) (*protocol.ShowDocumentResult, error) {
+func (cli *client) ShowDocument(ctx context.Context, params *protocol.ShowDocumentParams) (*protocol.ShowDocumentResult, error) {
 	var success bool
 	if params.External {
 		// Open URI in external browser.
 		success = browser.Open(params.URI)
 	} else {
 		// Open file in editor, optionally taking focus and selecting a range.
-		// (cmdClient has no editor. Should it fork+exec $EDITOR?)
+		// (client has no editor. Should it fork+exec $EDITOR?)
 		log.Printf("Server requested that client editor open %q (takeFocus=%t, selection=%+v)",
 			params.URI, params.TakeFocus, params.Selection)
 		success = true
@@ -645,33 +730,37 @@ func (c *cmdClient) ShowDocument(ctx context.Context, params *protocol.ShowDocum
 	return &protocol.ShowDocumentResult{Success: success}, nil
 }
 
-func (c *cmdClient) WorkDoneProgressCreate(context.Context, *protocol.WorkDoneProgressCreateParams) error {
+func (cli *client) WorkDoneProgressCreate(context.Context, *protocol.WorkDoneProgressCreateParams) error {
 	return nil
 }
 
-func (c *cmdClient) DiagnosticRefresh(context.Context) error {
+func (cli *client) DiagnosticRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) InlayHintRefresh(context.Context) error {
+func (cli *client) InlayHintRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) SemanticTokensRefresh(context.Context) error {
+func (cli *client) SemanticTokensRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) InlineValueRefresh(context.Context) error {
+func (cli *client) InlineValueRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) getFile(uri protocol.DocumentURI) *cmdFile {
-	file, found := c.files[uri]
+// getFile returns the specified file, adding it to the client state if needed.
+func (cli *client) getFile(uri protocol.DocumentURI) *cmdFile {
+	cli.filesMu.Lock()
+	defer cli.filesMu.Unlock()
+
+	file, found := cli.files[uri]
 	if !found || file.err != nil {
 		file = &cmdFile{
 			uri: uri,
 		}
-		c.files[uri] = file
+		cli.files[uri] = file
 	}
 	if file.mapper == nil {
 		content, err := os.ReadFile(uri.Path())
@@ -684,30 +773,38 @@ func (c *cmdClient) getFile(uri protocol.DocumentURI) *cmdFile {
 	return file
 }
 
-func (c *cmdClient) openFile(uri protocol.DocumentURI) *cmdFile {
-	c.filesMu.Lock()
-	defer c.filesMu.Unlock()
-	return c.getFile(uri)
-}
-
-// TODO(adonovan): provide convenience helpers to:
-// - map a (URI, protocol.Range) to a MappedRange;
-// - parse a command-line argument to a MappedRange.
-func (c *connection) openFile(ctx context.Context, uri protocol.DocumentURI) (*cmdFile, error) {
-	file := c.client.openFile(uri)
+// openFile returns the specified file, adding it to the client state
+// if needed, and notifying the server that it was opened.
+func (cli *client) openFile(ctx context.Context, uri protocol.DocumentURI) (*cmdFile, error) {
+	file := cli.getFile(uri)
 	if file.err != nil {
 		return nil, file.err
 	}
 
+	// Choose language ID from file extension.
+	var langID protocol.LanguageKind // "" eventually maps to file.UnknownKind
+	switch filepath.Ext(uri.Path()) {
+	case ".go":
+		langID = "go"
+	case ".mod":
+		langID = "go.mod"
+	case ".sum":
+		langID = "go.sum"
+	case ".work":
+		langID = "go.work"
+	case ".s":
+		langID = "go.s"
+	}
+
 	p := &protocol.DidOpenTextDocumentParams{
 		TextDocument: protocol.TextDocumentItem{
 			URI:        uri,
-			LanguageID: "go",
+			LanguageID: langID,
 			Version:    1,
 			Text:       string(file.mapper.Content),
 		},
 	}
-	if err := c.Server.DidOpen(ctx, p); err != nil {
+	if err := cli.server.DidOpen(ctx, p); err != nil {
 		// TODO(adonovan): is this assignment concurrency safe?
 		file.err = fmt.Errorf("%v: %v", uri, err)
 		return nil, file.err
@@ -715,38 +812,26 @@ func (c *connection) openFile(ctx context.Context, uri protocol.DocumentURI) (*c
 	return file, nil
 }
 
-func (c *connection) semanticTokens(ctx context.Context, p *protocol.SemanticTokensRangeParams) (*protocol.SemanticTokens, error) {
-	// use range to avoid limits on full
-	resp, err := c.Server.SemanticTokensRange(ctx, p)
-	if err != nil {
-		return nil, err
-	}
-	return resp, nil
-}
-
-func (c *connection) diagnoseFiles(ctx context.Context, files []protocol.DocumentURI) error {
-	cmd, err := command.NewDiagnoseFilesCommand("Diagnose files", command.DiagnoseFilesArgs{
+func diagnoseFiles(ctx context.Context, server protocol.Server, files []protocol.DocumentURI) error {
+	cmd := command.NewDiagnoseFilesCommand("Diagnose files", command.DiagnoseFilesArgs{
 		Files: files,
 	})
-	if err != nil {
-		return err
-	}
-	_, err = c.ExecuteCommand(ctx, &protocol.ExecuteCommandParams{
-		Command:   cmd.Command,
-		Arguments: cmd.Arguments,
-	})
+	_, err := executeCommand(ctx, server, cmd)
 	return err
 }
 
-func (c *connection) terminate(ctx context.Context) {
-	//TODO: do we need to handle errors on these calls?
-	c.Shutdown(ctx)
-	//TODO: right now calling exit terminates the process, we should rethink that
-	//server.Exit(ctx)
+func (cli *client) terminate(ctx context.Context) {
+	if err := cli.server.Shutdown(ctx); err != nil {
+		log.Printf("server shutdown failed: %v", err)
+	}
+
+	// Don't call Exit as it terminates the server process,
+	// which is the same as this client process.
+	// c.server.Exit(ctx)
 }
 
 // Implement io.Closer.
-func (c *cmdClient) Close() error {
+func (cli *client) Close() error {
 	return nil
 }
 
@@ -805,7 +890,7 @@ func (f *cmdFile) spanLocation(s span) (protocol.Location, error) {
 	if err != nil {
 		return protocol.Location{}, err
 	}
-	return f.mapper.RangeLocation(rng), nil
+	return f.mapper.URI.Location(rng), nil
 }
 
 // spanRange converts a (UTF-8) span to a protocol (UTF-16) range.
@@ -817,7 +902,7 @@ func (f *cmdFile) spanRange(s span) (protocol.Range, error) {
 	// case-sensitive directories. The authoritative answer
 	// requires querying the file system, and we don't want
 	// to do that.
-	if !strings.EqualFold(filepath.Base(string(f.mapper.URI)), filepath.Base(string(s.URI()))) {
+	if !strings.EqualFold(f.mapper.URI.Base(), s.URI().Base()) {
 		return protocol.Range{}, bugpkg.Errorf("mapper is for file %q instead of %q", f.mapper.URI, s.URI())
 	}
 	start, err := pointPosition(f.mapper, s.Start())
@@ -841,3 +926,17 @@ func pointPosition(m *protocol.Mapper, p point) (protocol.Position, error) {
 	}
 	return protocol.Position{}, fmt.Errorf("point has neither offset nor line/column")
 }
+
+// TODO(adonovan): delete in 2025.
+type fix struct{ app *Application }
+
+func (*fix) Name() string       { return "fix" }
+func (cmd *fix) Parent() string { return cmd.app.Name() }
+func (*fix) Usage() string      { return "" }
+func (*fix) ShortHelp() string  { return "apply suggested fixes (obsolete)" }
+func (*fix) DetailedHelp(flags *flag.FlagSet) {
+	fmt.Fprintf(flags.Output(), `No longer supported; use "gopls codeaction" instead.`)
+}
+func (*fix) Run(ctx context.Context, args ...string) error {
+	return tool.CommandLineErrorf(`no longer supported; use "gopls codeaction" instead`)
+}
